field_method_separator = '''"|"'''

from .constructor_content import constructor_content
from .field_content import field_content
from .method_content import method_content

# mari bikin interface, abstract class, dll di sini (sementara)
class_item = f"""
class_item: "@" class_config? class_name class_content

// public
// implements interface x
// extends class x
// export
// decorated: @InputType()
class_config: "/" classconfiglist "/"
classconfiglist: classconfig ("," classconfig)*

// @/i:observable,:parent_class,@override/myclass
// bisa mult interface: @/i:satu,i:dua/myclass
// bisa juga bikin utk bantuan ini dan itu: graphql? grpc? mvc model? db table?
classconfig: "+" -> public
  | "-" -> private    // perlu utk class?
  | "#" -> protected  // perlu utk class?
  | "i" -> interface  // interface declare
  | "i+" -> interface_add // buat class decl+interface decl
  | "t" -> type_alias // kita buat juga type_alias versi single statement
  | "t+" -> type_alias  // class+type
  | "@" nama_identifier   -> decorator
  | ":" nama_identifier   -> extends
  | "i:" nama_identifier  -> implements // interface (beda i dan i: = beda func decl dan call)
  | "^" -> abstract // juga perlu final dan strictfp
  | "%" -> static
  | "F" -> final
  | "SF" -> strictfp
  | "D" -> default
  | "A" -> async
  | ">" -> explicit_default_constructor // percepat ngetik

// masukkan initializer list di sini...

// class name hrs bisa terima type parameters...
class_name: nama_identifier_with_typeparams

class_content: "{{" classcontentlist* "}}"
classcontentlist: classcontent ({field_method_separator} classcontent)*
// field content
// method content
// field dan method terbedakan dg adanya funcargs-funcbody utk method
classcontent: field_content
  | constructor_content
  | method_content
  |

{field_content}

{method_content}

{constructor_content}
"""



from app.transpiler.zgenerate.refactor.handlers import (function_content,
                                                        function_param)
from app.transpiler.zgenerate.refactor.handlers.common import (dec, inc,
                                                               tabify_content)
from app.treeutils import (anak, beranak, chdata, child, child1, child2,
                           child3, child4, chtoken, data, ispohon, istoken,
                           jumlahanak, sebanyak, token)


def constructor_content(tree, structname=None, language="py"):
    kembali = ""

    ctorparam, ctorcontent = "", ""
    for item in anak(tree):
        if data(item) == "constructor_config":
            pass
        if data(item) == "function_param":
            ctorparam = function_param.function_param(item, language=language)
        if data(item) == "function_content":
            ctorcontent, jumlah_statement = function_content.function_content(
                item, language=language
            )
            if not jumlah_statement:
                ctorcontent = "{}"
            else:
                inc()
                # @my{>(){?+satu}}
                kiri = "{\n"
                kanan = "\n}"
                ctorcontent = kiri + tabify_content(ctorcontent) + kanan
                dec()

    kembali = f"{structname}{ctorparam} {ctorcontent}"
    inc()
    kembali = tabify_content(kembali)
    dec()

    if language == "py":
        pass
    elif language == "ts":
        pass
    elif language == "rs":
        pass
    elif language == "java":
        pass
    elif language == "kt":
        pass
    elif language == "go":
        pass
    elif language == "rb":
        pass
    elif language == "v":
        pass
    elif language == "dart":
        pass
    elif language == "cpp":
        pass
    elif language == "cs":
        pass
    elif language == "hs":
        pass
    elif language == "clj":
        pass
    elif language == "scala":
        pass
    elif language == "php":
        pass
    elif language == "swift":
        pass
    elif language == "elixir":
        pass
    elif language == "erlang":
        pass
    return kembali



# identifier?u//perintah
gui_operation = """
gui_operation: penanda "u" gui_operation_config? gui_operation_cmd
gui_operation_cmd: HURUF_DIGIT

gui_operation_config: "/" gui_operation_config_items "/"
gui_operation_config_items: gui_operation_config_item ("," gui_operation_config_item)*
gui_operation_config_item: "c"
"""



from app.transpiler.zgenerate.refactor.handlers import (condition_body,
                                                        condition_for)
from app.transpiler.zgenerate.refactor.handlers.common import (dec, inc,
                                                               tabify_content)
from app.treeutils import (anak, beranak, chdata, child, child1, child2,
                           child3, child4, chtoken, data, ispohon, istoken,
                           jumlahanak, sebanyak, token)


def for_statement(tree, language="py"):
    kembali = ""

    kembali, forname, forcondition, forbody = "", "", "", ""
    for item in anak(tree):
        if data(item) == "for_keyword":
            forname = "for"
        elif data(item) == "condition_for":
            forcondition = condition_for.condition_for(item, language=language)
            # print('peroleh for condition', forcondition)
        elif data(item) == "condition_body":
            inc()
            forbody = condition_body.condition_body(item, language=language)
            if not forbody:
                forbody = "{}"
            else:
                forbody = tabify_content(forbody)
                forbody = "{\n" + forbody + "\n}"
            dec()
    # utk for traditional, hasil adlh arrayname.forEach(() => __BODY__)
    if "__BODY__" in forcondition:
        """
        for <-- forname gak perlu
        items.forEach((item, idx) => {
        __BODY__
        })
        {} <- condition_body replace __BODY__
        """
        kembali = forcondition.replace("__BODY__", forbody)
    else:
        kembali = f"{forname} {forcondition} {forbody}"

    if language == "py":
        pass
    elif language == "ts":
        pass
    elif language == "rs":
        pass
    elif language == "java":
        pass
    elif language == "kt":
        pass
    elif language == "go":
        pass
    elif language == "rb":
        pass
    elif language == "v":
        pass
    elif language == "dart":
        pass
    elif language == "cpp":
        pass
    elif language == "cs":
        pass
    elif language == "hs":
        pass
    elif language == "clj":
        pass
    elif language == "scala":
        pass
    elif language == "php":
        pass
    elif language == "swift":
        pass
    elif language == "elixir":
        pass
    elif language == "erlang":
        pass
    return kembali



from app.transpiler.zgenerate.refactor.handlers import (
    enum_declaration, expression_item, for_statement, if_statement,
    instantiation_expression, single_statement, switch_statement,
    while_statement)
from app.treeutils import (anak, beranak, chdata, child, child1, child2,
                           child3, child4, chtoken, data, ispohon, istoken,
                           jumlahanak, sebanyak, token)


def statement_item(tree, within_statement_list=False, language="py"):
    print("statement_item")
    kembali = ""
    statement = child1(tree)
    if data(statement) == "single_statement":
        kembali += single_statement.single_statement(statement, language=language)
    elif data(statement) == "instantiation_expression":
        kembali += instantiation_expression.instantiation_expression(
            statement, language=language
        )
    elif data(statement) == "for_statement":
        kembali += for_statement.for_statement(statement, language=language)
    elif data(statement) == "if_statement":
        kembali += if_statement.if_statement(statement, language=language)
    elif data(statement) == "expression_item":
        kembali += expression_item.expression_item(statement, language=language)
    elif data(statement) == "while_statement":
        kembali += while_statement.while_statement(statement, language=language)
    elif data(statement) == "switch_statement":
        kembali += switch_statement.switch_statement(statement, language=language)
    elif data(statement) == "enum_declaration":
        kembali += enum_declaration.enum_declaration(statement, language=language)

    if language == "py":
        pass
    elif language == "ts":
        # if not kembali.endswith(';') and not within_statement_list:
        if not kembali.endswith(";"):
            kembali += ";"
    elif language == "rs":
        pass
    elif language == "java":
        pass
    elif language == "kt":
        pass
    elif language == "go":
        pass
    elif language == "rb":
        pass
    elif language == "v":
        pass
    elif language == "dart":
        pass
    elif language == "cpp":
        pass
    elif language == "cs":
        pass
    elif language == "hs":
        pass
    elif language == "clj":
        pass
    elif language == "scala":
        pass
    elif language == "php":
        pass
    elif language == "swift":
        pass
    elif language == "elixir":
        pass
    elif language == "erlang":
        pass
    return kembali



# kita pengen buka langsung fmus
# !!N
# !!N#basedir
# !!N##filename
# !!N###filepath
# !!X#...fmus code...
fmus_ops = """
fmus_ops: fmus_keyword fmus_config? fmus_cmd

fmus_config: "/" fmus_configlist "/"
fmus_configlist: fmus_configitem ("," fmus_configitem)*
fmus_configitem: "x"

fmus_cmd: fmus_new
  | fmus_execute
  | fmus_run_file

fmus_new: "N" fmus_new_args?
fmus_new_args: "#" HURUF_FOLDER -> fmus_new_basedir
  | "##" HURUF_FOLDER -> fmus_new_filename // under /tmp
  | "###" HURUF_FOLDER -> fmus_new_filepath

fmus_execute: "X#" HURUF_FMUSCODE
fmus_run_file: "R#" HURUF_DIGIT_SPASI

HURUF_FOLDER:		("_"|LETTER|DIGIT|"/"|".") 	("_"|LETTER|DIGIT|"."|"-"|"/"|","|" "|"@"|"+")*
HURUF_FMUSCODE: ("_"|LETTER|DIGIT|"*"|"/"|"."|"$"|"\\""|"("|")"|"*"|"&") (LETTER|DIGIT|"$"|"_"|"*"|"."|","|"/"|"-"|"+"|"="|"&"|"@"|"%"|"("|")"|"["|"]"|"{"|"}"|" "|":"|";"|"\\\\"|"\\""|"!"|"<"|">"|"?"|"|"|"'"|"#"|"~"|"^")*
"""



dataframe_operation = """
dataframe_operation:    penanda "#" dataframe_operation_config? dataframe_operation_cmd
// sementara bebas krn gak tau batas2 operasinya
dataframe_operation_cmd: HURUF_DIGIT

dataframe_operation_config: "/" dataframe_operation_config_items "/"
dataframe_operation_config_items: dataframe_operation_config_item ("," dataframe_operation_config_item)*
dataframe_operation_config_item: "c"
"""



switch_statement = """
switch_statement: switch_keyword switch_config? condition_switch case_body+
switch_keyword: "/s"
  | "switch"
  | "sw"
  | "s"

condition_switch: "(" expression_item ")"

case_body: (condition_case|condition_defaultcase) condition_body

condition_case: "(" expression_item ")"
condition_defaultcase: "()"

switch_config: "/" switch_config_items "/"
switch_config_items: switch_config_item ("," switch_config_item)*
// gunakan switch(literal) dll
switch_config_item: "c"
"""



from app.transpiler.zgenerate.refactor.handlers import (class_item,
                                                        function_item,
                                                        interface_item)
from app.treeutils import (anak, beranak, chdata, child, child1, child2,
                           child3, child4, chtoken, data, ispohon, istoken,
                           jumlahanak, sebanyak, token)


def block_item(tree, language="py"):
    print("block_item")
    kembali = ""
    for cucu in anak(tree):
        if cucu.data == "function_item":
            kembali += function_item.function_item(cucu, language=language)
        elif cucu.data == "class_item":
            kembali += class_item.class_item(cucu, language=language)
        elif cucu.data == "interface_item":
            kembali += interface_item.interface_item(cucu, language=language)
    if language == "py":
        pass
    elif language == "ts":
        pass
    elif language == "rs":
        pass
    elif language == "java":
        pass
    elif language == "kt":
        pass
    elif language == "go":
        pass
    elif language == "rb":
        pass
    elif language == "v":
        pass
    elif language == "dart":
        pass
    elif language == "cpp":
        pass
    elif language == "cs":
        pass
    elif language == "hs":
        pass
    elif language == "clj":
        pass
    elif language == "scala":
        pass
    elif language == "php":
        pass
    elif language == "swift":
        pass
    elif language == "elixir":
        pass
    elif language == "erlang":
        pass
    return kembali



from app.transpiler.zgenerate.refactor.handlers import (
    class_config, class_content, nama_identifier_with_typeparams)
from app.transpiler.zgenerate.refactor.handlers.common import dec, inc
from app.treeutils import (anak, beranak, chdata, child, child1, child2,
                           child3, child4, chtoken, data, ispohon, istoken,
                           jumlahanak, sebanyak, token)


def class_item(tree, language="py"):
    kembali = ""

    class_modifier_start, class_modifier_end, class_name, _class_content = (
        "",
        "",
        "",
        "",
    )
    for item in anak(tree):
        if data(item) == "class_config":
            # harus cek jk interface atau interface_add maka proses class_item jadi berbeda
            class_modifier_start, class_modifier_end = class_config.class_config(
                item, language=language
            )
            # kembali += prepender + ' '
        elif data(item) == "class_name":
            # class_name
            if chdata(item) == "nama_identifier_with_typeparams":
                _nama_identifier_with_typeparams = child(item)
                class_name = (
                    nama_identifier_with_typeparams.nama_identifier_with_typeparams(
                        _nama_identifier_with_typeparams, language=language
                    )
                )
            elif chdata(item) == "nama_identifier":
                class_name = token(item)
            # kembali += f'class {namakelas} '
        elif data(item) == "class_content":
            # class_content
            inc()
            _class_content = class_content.class_content(
                item, structname=class_name, language=language
            )
            dec()

    if class_modifier_start:
        kembali += class_modifier_start + " "
    kembali += f"{class_name}"
    if class_modifier_end:
        kembali += " " + class_modifier_end
    if not _class_content:
        kembali += " {}"
    else:
        kembali += " {\n"
        kembali += _class_content
        kembali += "\n}"

    if language == "py":
        pass
    elif language == "ts":
        pass
    elif language == "rs":
        pass
    elif language == "java":
        pass
    elif language == "kt":
        pass
    elif language == "go":
        pass
    elif language == "rb":
        pass
    elif language == "v":
        pass
    elif language == "dart":
        pass
    elif language == "cpp":
        pass
    elif language == "cs":
        pass
    elif language == "hs":
        pass
    elif language == "clj":
        pass
    elif language == "scala":
        pass
    elif language == "php":
        pass
    elif language == "swift":
        pass
    elif language == "elixir":
        pass
    elif language == "erlang":
        pass
    return kembali



# from .zgenerate.gen_js import generate as generate_js
from .zgenerate.gen_dart import generate as generate_dart
from .zgenerate.gen_go import generate as generate_go
from .zgenerate.gen_java import generate as generate_java
from .zgenerate.gen_kt import generate as generate_kt
from .zgenerate.gen_py import generate as generate_py
from .zgenerate.gen_rb import generate as generate_rb
from .zgenerate.gen_rs import generate as generate_rs
from .zgenerate.gen_ts import generate as generate_ts
from .zgenerate.gen_v import generate as generate_v

languages = [
    "clj",
    "cpp",
    "cs",
    "dart",
    "elixir",
    "erlang",
    "go",
    "hs",
    "java",
    "js",
    "kt",
    "php",
    "pl",
    "py",
    "rb",
    "scala",
    "swift",
    "ts",
    "v",
]

daftar_languages = {
    "py": generate_py,
    "dart": generate_dart,
    "go": generate_go,
    "java": generate_java,
    "kt": generate_kt,
    "rb": generate_rb,
    "rs": generate_rs,
    "ts": generate_ts,
    "v": generate_v,
    # unused
    "clj": generate_py,
    "cpp": generate_py,
    "cs": generate_py,
    "elixir": generate_py,
    "erlang": generate_py,
    "hs": generate_py,
    "js": generate_py,
    "php": generate_py,
    "scala": generate_py,
}

program_context = {
    "language_context": [],
    "library_context": [],
    "framework_context": [],
    "tools_context": [],
}

scope_context = {}

statement_context = {}

library_context = {}

framework_context = {}

tools_context = {}



# /=satu,dua,tiga/empat/
# =/satu,dua,tiga/empat/
# =/o/satu,dua,tiga/empat/
destructuring_statement = """
destructuring_statement: destructuring_markstart destructuring_config? destructuring_content destructuring_markend
destructuring_markstart: "=/"
destructuring_markend: "/"
destructuring_config: destructuringconfiglist "/"
destructuringconfiglist: destructuringconfig+
destructuringconfig: "o" -> object // default
  | "l" -> list
  | "t" -> tuple

destructuring_content: destructuring_lhs "/" destructuring_rhs
destructuring_lhs: destructuring_lhs_item ("," destructuring_lhs_item)*

destructuring_lhs_item: HURUF_DIGIT_DESTRUCTURING
destructuring_rhs: HURUF_DIGIT_DESTRUCTURING
"""



from app.transpiler.zgenerate.refactor.handlers import enum_member
from app.transpiler.zgenerate.refactor.handlers.common import self_tab
from app.treeutils import (anak, beranak, chdata, child, child1, child2,
                           child3, child4, chtoken, data, ispohon, istoken,
                           jumlahanak, sebanyak, token)


def enum_body(tree, language="py"):
    kembali = ""

    memberlist = child1(tree)
    results = []
    for item in anak(memberlist):
        if data(item) == "enum_member":
            hasil = enum_member.enum_member(item, language=language)
            results.append(hasil)
    results = [self_tab() + item for item in results]
    kembali += ",\n".join(results)

    if language == "py":
        pass
    elif language == "ts":
        pass
    elif language == "rs":
        pass
    elif language == "java":
        pass
    elif language == "kt":
        pass
    elif language == "go":
        pass
    elif language == "rb":
        pass
    elif language == "v":
        pass
    elif language == "dart":
        pass
    elif language == "cpp":
        pass
    elif language == "cs":
        pass
    elif language == "hs":
        pass
    elif language == "clj":
        pass
    elif language == "scala":
        pass
    elif language == "php":
        pass
    elif language == "swift":
        pass
    elif language == "elixir":
        pass
    elif language == "erlang":
        pass
    return kembali



literal = """
literal: literal_number
  | literal_string
  | template_string // <my name is /name/ so take care yo>
  | literal_bool
  | literal_char
  | literal_list
  | literal_dict // termasuk literal obj di sini
  | literal_tuple // termasuk pair
  | literal_pair
  | literal_set

literal_template_string: literal_string
  | template_string

literal_number: BILBUL_BERTANDA
literal_string: "\\"" HURUF_TEMPLATESTRING "\\"" // "hello"
  | "'" HURUF_TEMPLATESTRING "'" // 'hello'
template_string: "<" HURUF_TEMPLATESTRING ">"
literal_bool: "T" -> boolean_true
  | "F" -> boolean_false
literal_char: "'" LETTER "'" // 'a'

literal_list: "[" list_items* "]"
list_items: list_item ("," list_item)*
list_item: expression_item

// krn () dan {} rebutan dg func param dan body
literal_dict: "{#" dict_items* "}"
dict_items: dict_item ("," dict_item)*
dict_item: dict_item_name ":" expression_item
dict_item_name: HURUF_DIGIT

literal_tuple: "(#" tuple_items* ")"
tuple_items: tuple_item ("," tuple_item)*
tuple_item: expression_item

literal_pair: "(@" tuple_item "," tuple_item ")"

// <> skrg dipake oleh template_string
literal_set: "<#" set_items* ">"
set_items: set_item ("," set_item)*
set_item: expression_item
"""



method_content = """
method_content: method_config? function_name tipe_identifier? function_param function_content

method_config: "/" methodconfiglist "/"
methodconfiglist: methodconfig ("," methodconfig)*
// decorated: @Override
// static
// public
methodconfig: "g" -> getter   // this method is a getter
  | "s" -> setter             // this method is a setter
  | "+" -> public
  | "-" -> private
  | "#" -> protected
  | "%" -> static     // ingat python ada class vs static yg berbeda dikit konsepnya
  | "A" -> async
  | "o" -> override   // cepat @Override etc
"""



# ini utk anonymous block dll
# ini cocok utk static block di ts: https://www.typescriptlang.org/docs/handbook/2/classes.html
# static {} dalam class
scope_item = """
// ||//{}

scope_item: keyword_scope scope_config? condition_body
//scope_item: keyword_scope scope_config? scope_content
//scope_content: "{" condition_body "}"

scope_config: "/" scopeconfiglist "/"
scopeconfiglist: scopeconfig ("," scopeconfig)*
scopeconfig: "sc" -> dummy_scope_config

//scope_content: "{" scopecontentlist "}"
//scopecontentlist: scopecontent ("," scopecontent)*
//scopecontent: "s"
//  |
"""



declaration_config = """

// mutable dan immutable sudah dihandle oleh $ dan %
// $ pasti mutable let mut myvar = value;
declaration_config: "/" declarationconfiglist "/"

//declaration_config: declarationconfiglist
declarationconfiglist: declare_config ("," declare_config)*

// instance field, static field
// $stack=[]
declare_config: "+" -> public
  | "-" -> private
  | "#" -> protected
  | "%" -> static
  | "L" -> late
  | "F" -> final

declaration_name: HURUF_DIGIT

declaration_value: "=" expression_item

"""



from app.transpiler.zgenerate.refactor.handlers import dataops_statement
from app.treeutils import (anak, beranak, chdata, child, child1, child2,
                           child3, child4, chtoken, data, ispohon, istoken,
                           jumlahanak, sebanyak, token)


def dataops_statement_with_identifier(tree, language="py"):
    kembali = ""

    identifier, rhs = "", ""
    for item in anak(tree):
        if data(item) == "nama_identifier":
            identifier = token(item)
        elif data(item) == "dataops_statement":
            rhs = dataops_statement.dataops_statement(
                item, identifier, language=language
            )

    if language == "py":
        if isinstance(rhs, dict):
            """
            kembalian anon func: dict berisi internal dan external
            """
            if rhs["type"] == "array_operation":
                operasi = rhs["operation"]
                if rhs["external"]:
                    kembali += f"{rhs['external']}\n"
                    kembali += "{operasi}({rhs['internal']}, {identifier})"
                else:
                    kembali = f"{operasi}({rhs['internal']}, {identifier})"
            elif rhs["type"] == "file_operation":
                kembali = f'const {identifier} = {rhs["result"]}'
        elif isinstance(rhs, str):
            kembali = rhs
    elif language == "ts":
        if isinstance(rhs, dict):
            """
            kembalian anon func: dict berisi internal dan external
            """
            if rhs["type"] == "array_operation":
                operasi = rhs["operation"]
                if rhs["external"]:
                    kembali += f"{rhs['external']}\n"
                    kembali += f"{identifier}.{operasi}({rhs['internal']})"
                else:
                    kembali = f"{identifier}.{operasi}({rhs['internal']})"
            elif rhs["type"] == "file_operation":
                kembali = f'const {identifier} = {rhs["result"]}'
        elif isinstance(rhs, str):
            kembali = rhs
    elif language == "rs":
        pass
    elif language == "java":
        pass
    elif language == "kt":
        pass
    elif language == "go":
        pass
    elif language == "rb":
        pass
    elif language == "v":
        pass
    elif language == "dart":
        pass
    elif language == "cpp":
        pass
    elif language == "cs":
        pass
    elif language == "hs":
        pass
    elif language == "clj":
        pass
    elif language == "scala":
        pass
    elif language == "php":
        pass
    elif language == "swift":
        pass
    elif language == "elixir":
        pass
    elif language == "erlang":
        pass
    return kembali



import datetime
from langutils.app.utils import trycopy
from . import process_language
from .lalang import shortcuts, bahasa



def myrepl():
    code = 1
    while code != "x":
        try:
            code = input(f'LALANG {datetime.datetime.now().strftime("%m/%d/%Y, %H:%M:%S")} >> ')
            code = code.strip()
            if code == "bahasa":
                print(bahasa)
                trycopy(bahasa)
            elif [item for item in shortcuts.keys() if item.startswith(code)]:
                # sementara blm pake args
                program = shortcuts[code]
                print(f"convert code from {code} to {program}")
                process_language(program)
            elif code != "" and code != "x":
                process_language(code)
        except EOFError as eof:
            print("Ctrl+D? Exiting...", eof)
            break
        except Exception as err:
            print(err)

if __name__ == "__main__":
    myrepl()



from app.transpiler.zgenerate.refactor.handlers import (anon_expression,
                                                        anon_statements,
                                                        function_config,
                                                        function_param)
from app.transpiler.zgenerate.refactor.handlers.common import (
    dec, inc, tabify_contentlist)
from app.treeutils import (anak, beranak, chdata, child, child1, child2,
                           child3, child4, chtoken, data, ispohon, istoken,
                           jumlahanak, sebanyak, token)


def anonymous_function(tree, language="py"):
    kembali = ""

    is_arrow_func_or_lambda = True  # jk hanya 1 expr
    funcconf, funcparams, funcbody = "", "", ""
    internal, external = "", ""
    for item in anak(tree):
        if data(item) == "non_arrow_func":
            is_arrow_func_or_lambda = False
        elif data(item) == "function_config":
            funcconf = function_config.function_config(item, language=language)
            funcconf = " ".join(funcconf)
        elif data(item) == "function_param":
            funcparams = function_param.function_param(item, language=language)
        elif data(item) == "anon_expression":
            hasil = anon_expression.anon_expression(item, language=language)
            internal = f"lambda item: {hasil}"
        elif data(item) == "anon_statements":
            badan_fungsi = anon_statements.anon_statements(item, language=language)
            # print('anon_statements #1:', badan_fungsi)
            if (
                len(badan_fungsi) == 1 and is_arrow_func_or_lambda
            ):  # 1 statement -> arrow
                # print('anon_statements #1a:')
                hasil = badan_fungsi[0].removesuffix(";")  # 1 expression jgn pake ;
                # print('anon_statements #2a:')
                if funcconf:
                    # print('anon_statements #3a:')
                    internal += funcconf + " "
                # print('anon_statements #4a:')
                internal += f"{funcparams} => {hasil}"
                # print('anon_statements #5a:')
            elif is_arrow_func_or_lambda:  # minta arrow -> arrow
                # print('anon_statements #1b:')
                inc()  # hrs sblm proses tabify
                funcbody = tabify_contentlist(badan_fungsi)
                dec()
                # if funcparams == '()':
                #   funcparams = '(item, index)'
                if funcconf:
                    internal += funcconf + " "
                internal += f"{funcparams} => {{\n{funcbody}\n}}"
            else:  # minta func
                # assign_name = 'anon_func'
                # internal = assign_name
                # print('anon_statements #1c:')
                inc()  # hrs sblm proses tabify
                funcbody = tabify_contentlist(badan_fungsi)
                dec()
                # if funcparams == '()':
                #   funcparams = '(item, index)'
                # external = f'def {assign_name}{funcparams}:\n' + funcbody + '\n'
                if funcconf:
                    internal += funcconf + " "
                internal += f"function {funcparams} {{\n" + funcbody + "\n}"

    # external adlh: def myanon_func(): ...
    # internal adlh: map/filter/reduce(myanon_func, mylist)
    kembali = {
        "internal": internal,
        "external": external,
    }

    if language == "py":
        pass
    elif language == "ts":
        pass
    elif language == "rs":
        pass
    elif language == "java":
        pass
    elif language == "kt":
        pass
    elif language == "go":
        pass
    elif language == "rb":
        pass
    elif language == "v":
        pass
    elif language == "dart":
        pass
    elif language == "cpp":
        pass
    elif language == "cs":
        pass
    elif language == "hs":
        pass
    elif language == "clj":
        pass
    elif language == "scala":
        pass
    elif language == "php":
        pass
    elif language == "swift":
        pass
    elif language == "elixir":
        pass
    elif language == "erlang":
        pass
    return kembali



huruf = """

HURUF_DIGIT_ONLY:       ("_"|LETTER|DIGIT) 	("_"|LETTER|DIGIT)*
HURUF_DIGIT: 						("_"|LETTER|DIGIT) 	("_"|LETTER|DIGIT|".")*
HURUF_EXPORT: 					("_"|LETTER|DIGIT|"*") 	("_"|LETTER|DIGIT|"."|"/"|"*")*
HURUF_DIGIT_DOT:				("_"|LETTER|DIGIT|".") 	("_"|LETTER|DIGIT|"."|"@"|"/")*
HURUF_DIGIT_SPASI: 			("_"|LETTER|DIGIT|"*") 	("_"|LETTER|DIGIT|"."|"@"|" "|"*")*
HURUF_DIGIT_MINUSPLUS: 	("_"|LETTER|DIGIT) 	("_"|LETTER|DIGIT|"."|"-"|"+"|"@")*
HURUF_TEMPLATESTRING:   ("_"|LETTER|DIGIT|"/") 	("_"|LETTER|DIGIT|" "|"."|","|"-"|"+"|"/"|"@")*
HURUF_SPACE: 						("_"|LETTER) 		("_"|LETTER|DIGIT|" ")*
HURUF_COMMA: 						("_"|LETTER) 		("_"|LETTER|DIGIT|",")*
HURUF_SYSTEM: 					("_"|LETTER|DIGIT|"*"|"/"|"."|"\\"") 	(LETTER|DIGIT|"_"|"*"|"."|"/"|"-"|"+"|" "|":"|"\\"|"\"")*
HURUF_PASSWORD: 				("_"|LETTER) ("_"|LETTER|DIGIT|" "|"&"|"%%"|"."|","|"-"|"_"|"+"|"@"|"#"|"*")*
HURUF_HOST: 						("_"|LETTER|DIGIT) ("_"|LETTER|DIGIT|"."|"@")*

HURUF_DIGIT_DESTRUCTURING: ("_"|LETTER|DIGIT) 	("_"|LETTER|DIGIT|".")*
"""



expression_item_header = """
expression_item: function_call  // myfunc(34), myfunc(param1=42)
  | anonymous_function          // (req,res){}
  | literal
  | member_dot_expression
  | member_index_expression
  | casting_expression
  | expression_with_operator
  | nama_identifier
  | new_operator nama_identifier function_call_param -> instantiation_expression
  | range_expression

casting_expression: expression_item (keyword_casting expression_item)+

expression_with_operator: "!" expression_item -> not_expression
  | "++" expression_item -> pre_inc_expression
  | "--" expression_item -> pre_dec_expression
  | expression_item "++" -> post_inc_expression
  | expression_item "--" -> post_dec_expression
  | expression_item "+=" literal -> plus_equal_expression
  | expression_item "-=" literal -> minus_equal_expression
  | expression_item "?" expression_item ":" expression_item -> ternary_expression
  | expression_item "&&" expression_item -> and_expression
  | expression_item "||" expression_item -> or_expression
  | expression_item arithmetic_operator expression_item -> arithmetic_expression
  | expression_item relational_operator expression_item -> relational_expression

// member index = obj[index]
// member dot = obj.member

// [ei]([ei])* atau ([ei])+
//member_index_expression: expression_item "[" expression_item "]" ("[" expression_item "]")*
member_index_expression: expression_item ("[" expression_item "]")+

// nama.nama(.nama)* atau nama (.nama)+
//member_dot_expression: nama_tanpa_dot "." nama_tanpa_dot ("." nama_tanpa_dot)*
member_dot_expression: nama_tanpa_dot ("." nama_tanpa_dot)+

"""
from .function_call import function_call
from .literal import literal
from .operator import operator
from .range_expression import range_expression

expression_item = f"""
{expression_item_header}

{function_call}

{operator}

{literal}

{range_expression}
"""



from app.transpiler.zgenerate.refactor.handlers import (literal_bool,
                                                        literal_char,
                                                        literal_dict,
                                                        literal_list,
                                                        literal_number,
                                                        literal_string,
                                                        template_string)
from app.treeutils import (anak, beranak, chdata, child, child1, child2,
                           child3, child4, chtoken, data, ispohon, istoken,
                           jumlahanak, sebanyak, token)


def literal(tree, language="py"):
    kembali = ""
    print("literal")
    for item in anak(tree):
        jenis = data(item)
        if jenis == "literal_string":
            kembali += literal_string.literal_string(item, language=language)
        elif jenis == "literal_char":
            kembali += literal_char.literal_char(item, language=language)
        elif jenis == "literal_number":
            kembali += literal_number.literal_number(item, language=language)
        elif jenis == "literal_list":
            hasil_list = literal_list.literal_list(item, language=language)
            kembali += str(hasil_list).replace("'", "")
        elif jenis == "literal_dict":
            kembali += literal_dict.literal_dict(item, language=language)
        elif jenis == "template_string":
            kembali += template_string.template_string(item, language=language)
        elif jenis.startswith("boolean"):
            kembali += literal_bool.literal_bool(item, language=language)

    if language == "py":
        pass
    elif language == "ts":
        pass
    elif language == "rs":
        pass
    elif language == "java":
        pass
    elif language == "kt":
        pass
    elif language == "go":
        pass
    elif language == "rb":
        pass
    elif language == "v":
        pass
    elif language == "dart":
        pass
    elif language == "cpp":
        pass
    elif language == "cs":
        pass
    elif language == "hs":
        pass
    elif language == "clj":
        pass
    elif language == "scala":
        pass
    elif language == "php":
        pass
    elif language == "swift":
        pass
    elif language == "elixir":
        pass
    elif language == "erlang":
        pass
    return kembali



from app.dirutils import joiner, new_filename_timestamp, tempdir
from app.transpiler.editor import editor
from app.transpiler.zgenerate.refactor.handlers.fmus_run import fmus_run
from app.treeutils import (anak, beranak, chdata, child, child1, child2,
                           child3, child4, chtoken, data, ispohon, istoken,
                           jumlahanak, sebanyak, token)


def fmus_cmd(tree, language="py"):
    kembali = ""

    code = ""
    filepath = joiner(tempdir(), new_filename_timestamp())

    for item in anak(tree):
        if data(item) == "fmus_execute":
            code = token(item)
            program = code + "\n"
            # print('executing program:', program)
            fmus_run(program, language=language)
            # print('selesai')
        elif data(item) == "fmus_new":
            for basefile in anak(item):
                if data(item) == "fmus_new_basedir":
                    basedir = token(item)
                    filepath = joiner(basedir, new_filename_timestamp())
                elif data(item) == "fmus_new_filename":
                    filename = token(item)
                    filepath = joiner(tempdir(), filename)
                elif data(item) == "fmus_new_filepath":
                    filepath = token(item)
            # pastikan filepath ada dan basedir terbuat
            program = editor(filepath)
            program = program + "\n"
            print("************")
            print(program)
            print("************")
            fmus_run(program, language=language)
    return filepath  # sementara...harusnya showtext/file utk coding

    if language == "py":
        pass
    elif language == "ts":
        pass
    elif language == "rs":
        pass
    elif language == "java":
        pass
    elif language == "kt":
        pass
    elif language == "go":
        pass
    elif language == "rb":
        pass
    elif language == "v":
        pass
    elif language == "dart":
        pass
    elif language == "cpp":
        pass
    elif language == "cs":
        pass
    elif language == "hs":
        pass
    elif language == "clj":
        pass
    elif language == "scala":
        pass
    elif language == "php":
        pass
    elif language == "swift":
        pass
    elif language == "elixir":
        pass
    elif language == "erlang":
        pass
    return kembali



from app.transpiler.zgenerate.refactor.handlers import literal
from app.treeutils import (anak, beranak, chdata, child, child1, child2,
                           child3, child4, chtoken, data, ispohon, istoken,
                           jumlahanak, sebanyak, token)


def function_call_param(tree, language="py"):
    kembali = ""
    if not beranak(tree):
        return "()"
    # callparamlist
    funcparamlist = child1(tree)
    if not beranak(funcparamlist):
        return "()"

    actualargs = []
    # print('iterate funcparamlist.children:', funcparamlist.children)
    for param in anak(funcparamlist):
        # callparam
        # print('func call param:', param)
        # nama_identifier or named_values
        if not beranak(param):
            continue
        namaid = child1(param)
        if data(namaid) == "nama_identifier":
            # print('param nama_identifier')
            argid = token(namaid)
            actualargs.append(argid)
        elif data(namaid) == "named_values":
            # print('param named_values')
            for kv in anak(namaid):
                # named_value = kv
                namaidentifier = str(kv.children[0].children[0])
                nilaiidentifier = kv.children[1]
                ei = nilaiidentifier.children[0]
                lit = ei.children[0]
                nilaiidentifierstr = literal.literal(lit, language=language)
                actualargs.append(f"{namaidentifier}={nilaiidentifierstr}")
        elif data(namaid) == "literal":
            hasil = literal.literal(namaid, language=language)
            actualargs.append(hasil)

    the_params = ", ".join(actualargs)

    kembali = f"({the_params})"

    if language == "py":
        pass
    elif language == "ts":
        pass
    elif language == "rs":
        pass
    elif language == "java":
        pass
    elif language == "kt":
        pass
    elif language == "go":
        pass
    elif language == "rb":
        pass
    elif language == "v":
        pass
    elif language == "dart":
        pass
    elif language == "cpp":
        pass
    elif language == "cs":
        pass
    elif language == "hs":
        pass
    elif language == "clj":
        pass
    elif language == "scala":
        pass
    elif language == "php":
        pass
    elif language == "swift":
        pass
    elif language == "elixir":
        pass
    elif language == "erlang":
        pass
    return kembali



# statement_separator = """
# statement_separator: ";"
# """

statement_list = f"""
statement_list: statement_item (statement_separator statement_item)*
"""

# condition_body: "{" single_statement* "}"

condition_body = f"""
{statement_list}

condition_body: "{{" statement_list* "}}"
"""



from app.transpiler.zgenerate.refactor.handlers import expression_item
from app.treeutils import (anak, beranak, chdata, child, child1, child2,
                           child3, child4, chtoken, data, ispohon, istoken,
                           jumlahanak, sebanyak, token)


def set_operation(tree, identifier, language="py"):
    kembali = ""

    for item in anak(tree):
        if data(item) == "penanda":
            pass
        elif data(item) == "insert_at_tail":  # ?S+
            ei = child1(item)
            hasil = expression_item.expression_item(ei, language=language)
            kembali += f"{identifier}.add({hasil})"
        elif data(item) == "concat_extend_update":  # ?S+=
            ei = child1(item)
            hasil = expression_item.expression_item(ei, language=language)
            kembali += f"{identifier}.update({hasil})"
        elif data(item) == "item_pop":  # ?S/
            kembali += f"{identifier}.pop()"
        elif data(item) == "item_pop_but_stay":  # ?S//
            # https://stackoverflow.com/questions/59825/how-to-retrieve-an-element-from-a-set-without-removing-it
            kembali += f"next(iter({identifier}))"
        elif data(item) == "length":
            kembali += f"len({identifier})"
        elif data(item) == "contains":
            ei = child1(item)
            hasil = expression_item.expression_item(ei, language=language)
            # kembali += f'{hasil} in {identifier}'
            kembali += f"{identifier}.includes({hasil})"
        elif data(item) == "count":
            ei = child1(item)
            hasil = expression_item.expression_item(ei, language=language)
            kembali += f"{identifier}.count({hasil})"
        elif data(item) == "clear":
            kembali += f"{identifier}.clear()"
        elif data(item) == "is_empty":
            kembali += identifier
        elif data(item) == "remove_item":
            ei = child1(item)
            hasil = expression_item.expression_item(ei, language=language)
            kembali += f"{identifier}.remove({hasil})"
        elif data(item) == "remove_item_silent":
            ei = child1(item)
            hasil = expression_item.expression_item(ei, language=language)
            kembali += f"{identifier}.discard({hasil})"
        elif data(item) == "to_string":
            kembali += f"str({identifier})"

        elif data(item) == "union":
            """
            https://www.w3schools.com/python/ref_set_union.asp
            set.union(set1, set2...)
            The union() method returns a set that contains all items from the original set, and all items from the specified set(s).
            You can specify as many sets you want, separated by commas.
            It does not have to be a set, it can be any iterable object.
            If an item is present in more than one set, the result will contain only one appearance of this item.
            """
            ei = child1(item)
            hasil = expression_item.expression_item(ei, language=language)
            kembali += f"{identifier}.union({hasil})"
        elif data(item) == "intersection":
            """
            https://www.w3schools.com/python/ref_set_intersection.asp
            set.intersection(set1, set2 ... etc)
            The intersection() method returns a set that contains the similarity between two or more sets.
            Meaning: The returned set contains only items that exist in both sets, or in all sets if the comparison is done with more than two sets.
            """
            ei = child1(item)
            hasil = expression_item.expression_item(ei, language=language)
            kembali += f"{identifier}.intersection({hasil})"
        elif data(item) == "difference":
            """
            set.difference(set)
            https://www.w3schools.com/python/ref_set_difference.asp
            The difference() method returns a set that contains the difference between two sets.
            Meaning: The returned set contains items that exist only in the first set, and not in both sets.
            """
            ei = child1(item)
            hasil = expression_item.expression_item(ei, language=language)
            kembali += f"{identifier}.difference({hasil})"
        elif data(item) == "symmetric_difference":
            """
            https://www.w3schools.com/python/ref_set_symmetric_difference.asp
            set.symmetric_difference(set)
            The symmetric_difference() method returns a set that contains all items from both set, but not the items that are present in both sets.
            Meaning: The returned set contains a mix of items that are not present in both sets.
            """
            ei = child1(item)
            hasil = expression_item.expression_item(ei, language=language)
            kembali += f"{identifier}.symmetric_difference({hasil})"
        elif data(item) == "is_disjoint":
            """
            gak ada persamaan sama sekali...
            The isdisjoint() method returns True if none of the items are present in both sets, otherwise it returns False.
            https://www.w3schools.com/python/ref_set_isdisjoint.asp
            set.isdisjoint(set)
            """
            ei = child1(item)
            hasil = expression_item.expression_item(ei, language=language)
            kembali += f"{identifier}.isdisjoint({hasil})"
        elif data(item) == "is_subset":
            """
            https://www.w3schools.com/python/ref_set_issubset.asp
            set.issubset(set)
            The issubset() method returns True if all items in the set exists in the specified set, otherwise it retuns False.
            """
            ei = child1(item)
            hasil = expression_item.expression_item(ei, language=language)
            kembali += f"{identifier}.issubset({hasil})"
        elif data(item) == "is_superset":
            """
            https://www.w3schools.com/python/ref_set_issuperset.asp
            set.issuperset(set)
            The issuperset() method returns True if all items in the specified set exists in the original set, otherwise it retuns False.
            """
            ei = child1(item)
            hasil = expression_item.expression_item(ei, language=language)
            kembali += f"{identifier}.issuperset({hasil})"

    if language == "py":
        pass
    elif language == "ts":
        pass
    elif language == "rs":
        pass
    elif language == "java":
        pass
    elif language == "kt":
        pass
    elif language == "go":
        pass
    elif language == "rb":
        pass
    elif language == "v":
        pass
    elif language == "dart":
        pass
    elif language == "cpp":
        pass
    elif language == "cs":
        pass
    elif language == "hs":
        pass
    elif language == "clj":
        pass
    elif language == "scala":
        pass
    elif language == "php":
        pass
    elif language == "swift":
        pass
    elif language == "elixir":
        pass
    elif language == "erlang":
        pass
    return kembali



# kita pengen dg gampang oprek redis...ini sbg ganti redis language
# rdbms -> isud + create db/user/role/table/view/index + alter table/column
# redis op: string, list, hash/dict, set
orm_operation = """
orm_operation:          penanda "O" orm_operation_config orm_operation_choice

// ?O/h=...,p=...,u=...,passwd=...,d=...,s=.../
orm_operation_config: "/" orm_operation_config_items "/"
orm_operation_config_items: orm_operation_config_item ("," orm_operation_config_item)*
orm_operation_config_item: "h=" HURUF_DIGIT -> orm_host_config
  | "p=" HURUF_DIGIT -> orm_port_config
  | "d=" HURUF_DIGIT -> orm_dbname_config
  | "s=" HURUF_DIGIT -> orm_schema_config
  | "u=" HURUF_DIGIT -> orm_username_config
  | "pw=" HURUF_DIGIT -> orm_password_config

// operasi spt Model.objects.all(), .get(id=...), .create(a=1,b=2)
orm_operation_choice: "s" -> orm_select        // ?Os
  | "f1" -> orm_find_one
  | "f1u" -> orm_find_one_and_update
  | "fn" -> orm_find_many
  | "fa" -> orm_find_all
  | "i" -> orm_insert        // ?Oi
  | "u" -> orm_update        // ?Ou
  | "d" -> orm_delete        // ?Od
  | "cd" -> orm_create_database
  | "ct" -> orm_create_table
  | "at" -> orm_alter_table
  | "ac" -> orm_alter_column
  | "cu" -> orm_create_user
  | "cr" -> orm_create_role
  | "ci" -> orm_create_index
  | "cv" -> orm_create_view
  | "cp" -> orm_create_procedure
  | "cg" -> orm_create_trigger
"""



from app.transpiler.zgenerate.refactor.handlers import (object_type,
                                                        tipe_identifier_array)
from app.transpiler.zgenerate.refactor.handlers.common import peta_tipe_data
from app.treeutils import (anak, beranak, chdata, child, child1, child2,
                           child3, child4, chtoken, data, ispohon, istoken,
                           jumlahanak, sebanyak, token)


def tipe_data_semua(tree, language="py"):
    kembali = ""

    for item in anak(tree):
        if data(item) == "array":
            kembali = tipe_identifier_array.tipe_identifier_array(
                item, language=language
            )
        elif data(item) == "object_type":
            kembali = object_type.object_type(item, language=language)
        else:
            jenis = data(item)
            kembali = peta_tipe_data[language].get(jenis, jenis)
    if language == "py":
        pass
    elif language == "ts":
        pass
    elif language == "rs":
        pass
    elif language == "java":
        pass
    elif language == "kt":
        pass
    elif language == "go":
        pass
    elif language == "rb":
        pass
    elif language == "v":
        pass
    elif language == "dart":
        pass
    elif language == "cpp":
        pass
    elif language == "cs":
        pass
    elif language == "hs":
        pass
    elif language == "clj":
        pass
    elif language == "scala":
        pass
    elif language == "php":
        pass
    elif language == "swift":
        pass
    elif language == "elixir":
        pass
    elif language == "erlang":
        pass
    return kembali



from app.transpiler.zgenerate.refactor.handlers.block_item import block_item
from app.transpiler.zgenerate.refactor.handlers.common import self_indentno
from app.transpiler.zgenerate.refactor.handlers.export_item import export_item
from app.transpiler.zgenerate.refactor.handlers.import_item import import_item
from app.transpiler.zgenerate.refactor.handlers.statement_item import \
    statement_item
from app.treeutils import (anak, beranak, chdata, child, child1, child2,
                           child3, child4, chtoken, data, ispohon, istoken,
                           jumlahanak, sebanyak, token)


def do_generate(RootNode, program_context, language):
    global self_indentno
    self_indentno = 0
    kembali = ""
    if data(RootNode) == "item":
        for dahan in anak(RootNode):
            if dahan.data == "block_item":
                kembali += block_item(dahan, language=language)
            elif dahan.data == "statement_item":
                kembali += statement_item(dahan, language=language)
            elif data(dahan) == "import_item":
                kembali += import_item(dahan, language=language)
            elif data(dahan) == "export_item":
                kembali += export_item(dahan, language=language)

    elif data(RootNode) == "item_separator":
        if RootNode.children:  # jk bukan []
            newline_or_not = chdata(RootNode)
            kembali += f"\tSEP: {newline_or_not}\n"

    return kembali


def generate(RootNode, program_context={}, language="ts"):
    return do_generate(RootNode, program_context, language=language)



from app.transpiler.zgenerate.refactor.handlers import (
    anonymous_function, arithmetic_expression, casting_expression,
    function_call, literal, member_dot_expression, member_index_expression,
    range_expression, relational_expression, ternary_expression)
from app.treeutils import (anak, beranak, chdata, child, child1, child2,
                           child3, child4, chtoken, data, ispohon, istoken,
                           jumlahanak, sebanyak, token)


def expression_item(tree, language="py"):
    kembali = ""
    provided_expression = child1(tree)
    hasil = ""
    if data(provided_expression) == "literal":
        hasil = literal.literal(provided_expression, language=language)
        # print('ei literal', type(literal), type(provided_expression), type(language), '=>', hasil)
    elif data(provided_expression) == "nama_identifier":
        hasil = token(provided_expression)
        # print('ei nama identifier:', hasil)
    elif data(provided_expression) == "relational_expression":
        hasil = relational_expression.relational_expression(
            provided_expression, language=language
        )
        # print('ei relational_expression', type(literal), type(provided_expression), type(language), '=>', hasil)
    elif data(provided_expression) == "function_call":
        hasil = function_call.function_call(provided_expression, language=language)
    elif data(provided_expression) == "arithmetic_expression":
        hasil = arithmetic_expression.arithmetic_expression(
            provided_expression, language=language
        )
    elif data(provided_expression) == "casting_expression":
        hasil = casting_expression.casting_expression(
            provided_expression, language=language
        )
    elif data(provided_expression) == "range_expression":
        hasil = range_expression.range_expression(
            provided_expression, language=language
        )
    elif data(provided_expression) == "member_index_expression":
        hasil = member_index_expression.member_index_expression(
            provided_expression, language=language
        )
    elif data(provided_expression) == "member_dot_expression":
        hasil = member_dot_expression.member_dot_expression(
            provided_expression, language=language
        )
    elif data(provided_expression) == "ternary_expression":
        # print('panggil ternary_expression dari expression_item')
        hasil = ternary_expression.ternary_expression(
            provided_expression, language=language
        )
    elif data(provided_expression) == "anonymous_function":
        hasil_in_dict = anonymous_function.anonymous_function(
            provided_expression, language=language
        )
        hasil = (
            hasil_in_dict["internal"]
            if hasil_in_dict["internal"]
            else hasil_in_dict["external"]
        )

    kembali += hasil

    if language == "py":
        pass
    elif language == "ts":
        pass
    elif language == "rs":
        pass
    elif language == "java":
        pass
    elif language == "kt":
        pass
    elif language == "go":
        pass
    elif language == "rb":
        pass
    elif language == "v":
        pass
    elif language == "dart":
        pass
    elif language == "cpp":
        pass
    elif language == "cs":
        pass
    elif language == "hs":
        pass
    elif language == "clj":
        pass
    elif language == "scala":
        pass
    elif language == "php":
        pass
    elif language == "swift":
        pass
    elif language == "elixir":
        pass
    elif language == "erlang":
        pass
    return kembali



# ?r
redis_operation = """
redis_operation: penanda "r" redis_operation_config? redis_cmd (redis_key ("=" redis_value)?)?

redis_operation_config: "/" redis_operation_config_items "/"
redis_operation_config_items: redis_operation_config_item ("," redis_operation_config_item)*
redis_operation_config_item: "h=" HURUF_DIGIT -> redis_host_config
  | "p=" HURUF_DIGIT -> redis_port_config
  | "d=" HURUF_DIGIT -> redis_dbname_config
  | "s=" HURUF_DIGIT -> redis_schema_config
  | "u=" HURUF_DIGIT -> redis_username_config
  | "pw=" HURUF_DIGIT -> redis_password_config

redis_cmd: "c" -> redis_connect       // ?r/d=7/c
  | "r" -> redis_reset                // ?rr          = reset connection to default
  | "?" -> redis_exists         // ?r?key       does redis key exist?
  | "=" -> redis_keys           // ?r=*  
  | "s=" -> redis_string_get    // ?rs=key
  | "s+" -> redis_string_add    // ?rs+key=value
  | "s-" -> redis_string_del    // ?rs-key
  | "A+" -> redis_list_add      // ?rA+
  | "A-" -> redis_list_del      // ?rA-
  | "A=" -> redis_list_get      // ?rA=
  | "D+" -> redis_hash_add      // ?rD+
  | "D-" -> redis_hash_del      // ?rD-
  | "D=" -> redis_hash_get      // ?rD=
  | "S+" -> redis_set_add       // ?rS+
  | "S-" -> redis_set_del       // ?rS-
  | "S=" -> redis_set_get       // ?rS=

redis_args: redis_arg ("," redis_arg)*
redis_arg: redis_key
  | redis_value

redis_key: HURUF_EXPORT
// sementara gini dulu...tapi nanti butuh value utk dict etc
redis_value: HURUF_EXPORT 
"""



keywords = """
keyword_function: ":"
keyword_import: "I"
keyword_package: "P"
keyword_export: "X:"
keyword_casting: ">>"
//keyword_scope: "(s)"
keyword_scope: "||"
//
range_keyword: ".."

typealias_keyword: "&&"
interface_keyword: "@@"
if_keyword: "if"
//searchable_kw: "~"
searchable_kw: "`" // `kata utk dicari`
fmus_keyword: "!!"
enum_keyword: "&"
exception_keyword: "x:"
csv_keyword: "csv:"
"""



set_operation = """
set_operation:        penanda "S" set_operation_choice
// lst?A
set_operation_choice: "/" -> item_pop
  | "//"                  -> item_pop_but_stay
  | "+=" expression_item  -> concat_extend_update
  | "+" expression_item   -> insert_at_tail
  | "|"                   -> length
  | "~" expression_item   -> contains
  | "0"                   -> clear
  | "#" expression_item   -> count
  | "0?"                  -> is_empty
  | "-" expression_item   -> remove_item
  | "--" expression_item  -> remove_item_silent
  | "2s"                  -> to_string

  | "U" expression_item                  -> union
  | "X" expression_item                  -> intersection
  | "<>" expression_item                 -> difference
  | "<<>>" expression_item               -> symmetric_difference
  | "-?" expression_item                 -> is_disjoint
  | "_?" expression_item                 -> is_subset
  | "^?" expression_item                 -> is_superset

"""



from declang.app.transpiler.frontend.bahasa import declarative_program, huruf_berbeda

from .assignment_statement import assignment_statement
from .const_declaration import const_declaration
from .declaration_config import declaration_config
from .destructuring_statement import destructuring_statement
from .typealias_declaration import typealias_declaration
from .var_declaration import var_declaration

single_statement = f"""
////////////////////////////////////////////////////////////////
// assignment
// initialization sama dg assignment?
// variable statement (dihandle $)
// typealias decl, enum decl
// throw (throw stmt), return (return stmt), yield (yield stmt), flow (continue+break stmt)
// with stmt, labelled stmt, debugger stmt
// empty stmt -> typescript hanya semicolon (terminator)
// abstract decl, namespace decl, interface decl, decorator list, generatorfunc decl
// expressionSequence adlh if(..), for(..; ..; ..), while(..), switch(..)
// tmsk di sini: function item (func+arrowfunc decl), exception item (try stmt), import item, export item, class item (class decl)
// tmsk: iteration stmt yg dihandle for+while
// tmsk: if stmt dan switch stmt yg sudah dihandle

single_statement: var_declaration
  | const_declaration
  | assignment_statement
  | flow_statement
  | return_statement
  | throw_statement
  | yield_statement
  | defer_statement
  | dataops_statement_with_identifier
  | stdout_operation
  | faker_ops
  | fmus_ops
  | redis_operation
  | typealias_declaration
  | destructuring_statement
  | "~" declarative_program

// dari go: defer statement

dataops_statement_with_identifier: nama_identifier dataops_statement

{assignment_statement}

{typealias_declaration}

flow_statement: flow_continue
  | flow_break
flow_continue: "continue"
  | "cont"
  | "co"
flow_break: "break"
  | "br"

return_statement: ">" expression_item
throw_statement: ">>" throw_name? expression_item
yield_statement: ">>>" expression_item
defer_statement: ">>>>" expression_item

throw_name: HURUF_DIGIT "/"

{destructuring_statement}

{var_declaration}

const_declaration: "%" declaration_config? declaration_name tipe_identifier? declaration_value?

{declaration_config}

declarative_program: HURUF_KODE_FRONTEND

{huruf_berbeda}
"""



from app.transpiler.zgenerate.refactor.handlers import tipe_identifier
from app.treeutils import (anak, beranak, chdata, child, child1, child2,
                           child3, child4, chtoken, data, ispohon, istoken,
                           jumlahanak, sebanyak, token)


def object_content(tree, language="py"):
    kembali = ""

    nama, jenis = "", ""
    for item in anak(tree):
        if data(item) == "nama_identifier":
            nama = token(item)
        elif data(item) == "tipe_identifier":
            jenis = tipe_identifier.tipe_identifier(item, language=language)
    kembali = f"{nama}: {jenis}"

    if language == "py":
        pass
    elif language == "ts":
        kembali += ";"
    elif language == "rs":
        pass
    elif language == "java":
        pass
    elif language == "kt":
        pass
    elif language == "go":
        pass
    elif language == "rb":
        pass
    elif language == "v":
        pass
    elif language == "dart":
        pass
    elif language == "cpp":
        pass
    elif language == "cs":
        pass
    elif language == "hs":
        pass
    elif language == "clj":
        pass
    elif language == "scala":
        pass
    elif language == "php":
        pass
    elif language == "swift":
        pass
    elif language == "elixir":
        pass
    elif language == "erlang":
        pass
    return kembali



from app.treeutils import (anak, beranak, chdata, child, child1, child2,
                           child3, child4, chtoken, data, ispohon, istoken,
                           jumlahanak, sebanyak, token)


def function_call_config(tree, language="py"):
    kembali = ""

    functionconfiglist = child1(tree)
    semua_modifier = []
    for nilai in anak(functionconfiglist):
        if data(nilai) == "async":
            hasil = "async"
            semua_modifier.append(hasil)
        elif data(nilai) == "await":
            hasil = "await"
            semua_modifier.append(hasil)

    if semua_modifier:
        kembali = " ".join(semua_modifier)
    else:
        kembali = ""

    if language == "py":
        pass
    elif language == "ts":
        pass
    elif language == "rs":
        pass
    elif language == "java":
        pass
    elif language == "kt":
        pass
    elif language == "go":
        pass
    elif language == "rb":
        pass
    elif language == "v":
        pass
    elif language == "dart":
        pass
    elif language == "cpp":
        pass
    elif language == "cs":
        pass
    elif language == "hs":
        pass
    elif language == "clj":
        pass
    elif language == "scala":
        pass
    elif language == "php":
        pass
    elif language == "swift":
        pass
    elif language == "elixir":
        pass
    elif language == "erlang":
        pass
    return kembali



# 10/random_int
# 10/random_int(0,100)
# 10/random_int(0,100)/A
faker_ops = """
faker_ops: penanda "K" faker_num "/" faker_cmd faker_args? as_output?
faker_num: BILBUL
faker_cmd: HURUF_DIGIT // nama fungsi
faker_args: "(" fake_arg ("," fake_arg)* ")"
fake_arg: HURUF_FUNCTIONCALL_ARGS
// default as_string utk 1 num, as_list utk multiple num
as_output: "/A" -> as_list
  | "/D" -> as_dict
  | "/S" -> as_set
  | "/L" -> as_tuple
  | "/P" -> as_pair
  | "/s" -> as_string
  | "/i" -> as_int

//HURUF_DIGIT: 						("_"|LETTER|DIGIT) 	("_"|LETTER|DIGIT|".")*
// ' hrs bisa mendahului spt 'int', 'str', utk literal string
// juga [] utk literal list
HURUF_FUNCTIONCALL_ARGS:   ("_"|LETTER|DIGIT|"'"|"["|"]") 	("_"|LETTER|DIGIT|" "|"."|"-"|"+"|"/"|"@"|"'"|"["|"]")*
"""



array_operation = """

array_operation:        penanda "A" array_operation_choice
// lst?A
array_operation_choice: "/" BILBUL -> item_at       // mylist?A/5 = get item at index
  | "/" BILBUL "=" expression_item -> item_at_set   // mylist[n] = value
  | "0/" -> item_first
  | "9/" -> item_last
  | "44" nama_identifier? -> array_traversal_for_traditional // myarr?A44
  | "4@" nama_identifier? -> array_traversal_for_each
  | "4#" nama_identifier? -> array_traversal_for_key
  | "4$" nama_identifier? -> array_traversal_for_value
  | "n" array_new_operation? -> new_array
  | "+=" expression_item  ("," expression_item)* -> concat_extend_update   // mylist?A+=   another list  (extend/update)
  | "+" expression_item   -> insert_at_tail         // mylist?A+    item          (insert_at_tail)
  | "+<" expression_item  -> insert_at_head
  | "+" expression_item "@" BILBUL -> insert_at_index
  | "|"                   -> length                 // mylist?A|
  | "#" expression_item   -> count
  | "i/" nama_identifier_or_literal -> index_of // mylist?Ai/5 = get index of item (item bisa id, atau literal spt 42 dan "hello")
  | "P" BILBUL?           -> pop_item       // remove at tail
  | "S" BILBUL?           -> shift_item     // remove at head
  | "X" BILBUL ("/" BILBUL)? -> slice_array
  | "-@" BILBUL           -> remove_at_index
  | "-<"                  -> remove_at_head
  | "->"                  -> remove_at_tail
  | "-" expression_item                 -> remove_by_value
  | "+S-" BILBUL "/" BILBUL "/" expression_item ("," expression_item)* -> splice_add_remove
  | "-S-" BILBUL "/" BILBUL -> splice_remove
  | "<<"                  -> min
  | ">>"                  -> max
  | "=="                  -> average
  | "++"                  -> sum
  | "~" expression_item   -> contains
  | "~~" expression_item  -> has_key // js dan dart, array bs punya key/index
  | "0"                   -> clear
  | "0?"                  -> is_empty         // mylist?A0?
  | "2s" -> to_string                         // ', '.join(mylist) = default by himpitkan
  | "2D" -> to_dict     // dart punya asMap(), mungkin java/kt jg
  | "sp" ("/" HURUF_SYSTEM "/")? -> split     // maksudnya join...
  | "!"                   -> reverse
  | "$"                   -> sort
  | "$!"                  -> reverse_sort
  | "M#" anonymous_function   -> array_map // mylist?AM#...
  | "F#" anonymous_function   -> array_filter
  | "R#" anonymous_function   -> array_reduce
  | "A?"                      -> is_array
  | "U"                       -> uniqify

// TODO: any(predicate), all(predicate)

array_new_operation: "@" BILBUL -> allocate // ?An@10, int[] intArray = new int[10];
  | "@@" BILBUL -> allocate_with_cap // ?An@10/1000
  | "@" -> just_create
  | "$" expression_item -> initialize // ?An$...ei...
  | "$" BILBUL tipe_identifier -> initialize_with_faker // ?An$5:s, ?An$5:i
  | "F" mapper_function_for_arraylike -> create_with_arrayfrom // ?AnF(){}, ?AnF:(){}
  | "@F" BILBUL "/" BILBUL -> allocate_with_arrayfrom // ?An@F5/0

mapper_function_for_arraylike: anonymous_function
"""



for_statement = """
for_statement: for_keyword for_config? condition_for condition_body

for_keyword: "/4"
  | "for"

// harus bisa handle: for, for each, for in, for of
// for i=0; i<expession; i++
// for-each item,key of array
// for-in key of array
// for-of item of array

condition_for: for_variation
for_variation: "(" for_traditional ")"  // for(i=0;i<dict.length;i++)
  |  "(" for_while? ")" -> for_ever // for(i<10), for()
  | "@" for_each "@"  // for@item/items/index@ = for key,value
  | "#" for_in "#"    // for#key/items# = for key/index
  | "$" for_of "$"    // for$item/items$ = for value

// apa gak mendingan: for(name,start,end,step) => for(index,0,42,2)
for_traditional: for_start ";" for_end (";" for_step)?
for_while: expression_item
for_each: item_name "/" array_name ("/" key_name)?
for_in: key_name "/" array_name
for_of: item_name "/" array_name

//for_start: key_name "=" expression_item
for_start: nama_jenis_identifier_optional "=" expression_item
for_end: expression_item
for_step: expression_item

key_name: nama_jenis_identifier_optional
item_name: nama_jenis_identifier_optional
array_name: nama_identifier
  | expression_item

for_config: "[" for_config_items "]"
for_config_items: for_config_item ("," for_config_item)*
// gunakan for-traditional, create local var, use nonlocal var, for-in, for-of
for_config_item: "c"
"""



from app.treeutils import (anak, beranak, chdata, child, child1, child2,
                           child3, child4, chtoken, data, ispohon, istoken,
                           jumlahanak, sebanyak, token)


def export_content(tree, language="py"):
    kembali = ""

    exported = []
    exportcontentlist = child1(tree)
    for item in anak(exportcontentlist):
        jenis = data(item)
        if jenis == "exportcontent":
            hasil = token(item)
            exported.append(hasil)

    kembali += ", ".join(exported)
    if language == "py":
        pass
    elif language == "ts":
        pass
    elif language == "rs":
        pass
    elif language == "java":
        pass
    elif language == "kt":
        pass
    elif language == "go":
        pass
    elif language == "rb":
        pass
    elif language == "v":
        pass
    elif language == "dart":
        pass
    elif language == "cpp":
        pass
    elif language == "cs":
        pass
    elif language == "hs":
        pass
    elif language == "clj":
        pass
    elif language == "scala":
        pass
    elif language == "php":
        pass
    elif language == "swift":
        pass
    elif language == "elixir":
        pass
    elif language == "erlang":
        pass
    return kembali



from app.transpiler.zgenerate.refactor.handlers import classcontent
from app.transpiler.zgenerate.refactor.handlers.common import self_tab
from app.treeutils import (anak, beranak, chdata, child, child1, child2,
                           child3, child4, chtoken, data, ispohon, istoken,
                           jumlahanak, sebanyak, token)


def classcontentlist(tree, structname=None, language="py"):
    kembali = ""

    collected_content = []
    for _classcontent in anak(tree):
        hasil = classcontent.classcontent(
            _classcontent, structname=structname, language=language
        )
        if hasil:
            collected_content.append(hasil)
    if collected_content:
        # tambah ; utk item
        tabbed_class_contents = [self_tab() + item + ";" for item in collected_content]
        tabbed_class_contents = "\n".join(tabbed_class_contents)
        kembali = tabbed_class_contents

    if language == "py":
        pass
    elif language == "ts":
        pass
    elif language == "rs":
        pass
    elif language == "java":
        pass
    elif language == "kt":
        pass
    elif language == "go":
        pass
    elif language == "rb":
        pass
    elif language == "v":
        pass
    elif language == "dart":
        pass
    elif language == "cpp":
        pass
    elif language == "cs":
        pass
    elif language == "hs":
        pass
    elif language == "clj":
        pass
    elif language == "scala":
        pass
    elif language == "php":
        pass
    elif language == "swift":
        pass
    elif language == "elixir":
        pass
    elif language == "erlang":
        pass
    return kembali



from app.transpiler.zgenerate.refactor.handlers import (declaration_value,
                                                        tipe_identifier)
from app.treeutils import (anak, beranak, chdata, child, child1, child2,
                           child3, child4, chtoken, data, ispohon, istoken,
                           jumlahanak, sebanyak, token)


def const_declaration(tree, language="py"):
    kembali = ""

    nama_const, jenis_const, nilai_const = "", "", ""
    for dahan in anak(tree):
        if data(dahan) == "declaration_config":
            pass
        elif data(dahan) == "declaration_name":
            nama_const = token(dahan)
        elif data(dahan) == "tipe_identifier":
            # jenis = data(child1(dahan))
            # tipe_native = peta_tipe_data.get(jenis, jenis)
            jenis_const = tipe_identifier.tipe_identifier(dahan, language=language)
            if jenis_const:
                jenis_const = ": " + jenis_const
        elif data(dahan) == "declaration_value":
            # ei = child1(dahan)
            # nilai_variable = expression_item(ei, returning=True)
            nilai_const = declaration_value.declaration_value(dahan, language=language)

    kembali = f"const {nama_const}{jenis_const} = {nilai_const}"

    if language == "py":
        pass
    elif language == "ts":
        kembali += ";"
    elif language == "rs":
        pass
    elif language == "java":
        pass
    elif language == "kt":
        pass
    elif language == "go":
        pass
    elif language == "rb":
        pass
    elif language == "v":
        pass
    elif language == "dart":
        pass
    elif language == "cpp":
        pass
    elif language == "cs":
        pass
    elif language == "hs":
        pass
    elif language == "clj":
        pass
    elif language == "scala":
        pass
    elif language == "php":
        pass
    elif language == "swift":
        pass
    elif language == "elixir":
        pass
    elif language == "erlang":
        pass
    return kembali



# identifier?@//perintah
react_operation = """
react_operation: penanda "@" react_operation_config? react_operation_cmd
react_operation_cmd: hook_operation
  | component_operation

// https://reactjs.org/docs/hooks-reference.html
hook_operation: "us" tipe_identifier? state_initializer? -> use_state
  | "ue" anonymous_function expression_item -> use_effect // useEffect(af, [ei])
  | "uc" nama_identifier "=" nama_identifier -> use_context
  | "ur" expression_item "=" expression_item -> use_reducer // const [state,dispatch]=uR(reducer,initState)
  | "uk" anonymous_function expression_item -> use_callback
  | "um" -> use_memo
  | "uR" -> use_ref
  | "uih" -> use_imperative_handle
  | "ule" -> use_layout_effect
  | "udv" -> use_debug_value
  | "gssp" -> get_server_side_props

state_initializer: "=" expression_item

component_operation: "fc" -> functional_component
  | "cc" -> class_component

react_operation_config: "/" react_operation_config_items "/"
react_operation_config_items: react_operation_config_item ("," react_operation_config_item)*
react_operation_config_item: "c"
"""
